/* The number plate game
 *
 * On ABC's 612 4QR Brisbane radio, Spencer Howsen plays a game
 * at about 3:30pm where he gives the caller 3 letters and they
 * have to come up with the longest word possible using those
 * three letters in that order.
 *
 * This is a program to.. well.. cheat.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <err.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#define PATH_WORDS	"/usr/share/dict/words"

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int fd;
	struct stat sb;
	const char *dictfile = PATH_WORDS;
	const char *matchword;
	const char *start, *end, *best, *word, *p, *m;
	int i, bestlen;
	char chmap[256];
	int ch, error = 0;
	extern char *optarg;
	extern int optind;

	while ((ch = getopt(argc, argv, "f:")) != -1)
		switch (ch) {
		case 'f':
			dictfile = optarg;
			break;
		default:
			error = 1;
			break;
		}
	if (argc != optind + 1)
		error = 1;
	if (error) {
		fprintf(stderr, "usage: %s [-f dictfile] matchword\n", argv[0]);
		exit(1);
	}

	matchword = argv[optind];
	if (!*matchword)
		errx(1, "empty matchword");

	/* Create a lowercase character map */
	for (i = 0; i < 256; i++)
		chmap[i] = (i >= 'A' && i <= 'Z') ? i - 'A' + 'a' : i;

	/* 'mmap' the file of words into memory for supreme speed */
	if ((fd = open(dictfile, O_RDONLY)) == -1)
		err(1, "open: %s", dictfile);
	if (fstat(fd, &sb) == -1)
		err(1, "fstat: %s", dictfile);
	start = (char *)mmap(NULL, sb.st_size, PROT_READ, MAP_FILE, fd, 0);
	if (start == (char *)MAP_FAILED)
		err(1, "%s: mmap", dictfile);
	end = start + sb.st_size;
	close(fd);

	/* Start matching */
	m = matchword;
	word = NULL;
	best = NULL;
	bestlen = 0;
	for (p = start; p < end; p++) {
		if (chmap[(int)*p] == chmap[(int)*m])
			m++;
		if (*p == '\n') {
			/* Check to see if we matched */
			if (*m == '\0' && (p - word) > bestlen) {
				best = word;
				bestlen = p - word;
				/* Print out best matches as we find them */
				printf("%.*s\n", bestlen, best);
			}
			m = matchword;
			word = p + 1;
		}
	}

	if (!best)
		errx(1, "no match");

	exit(0);
}
